"""
Data assimilation methods including ensemble simulation, satelite data retrival and processing and data assimilation algorithms
J. Fiddes, March 2022

TODO:
    - method to mosaic (gdal) multiple tiles before cropping

"""
import os
import glob
import re
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt
import copy
import rasterio as rio
from osgeo import gdal
import osr
from scipy.interpolate import interp1d
from TopoPyScale import topo_sim as sim
from datetime import datetime
import xarray as xr


# import gdal


def lognormDraws_kris(n, med, s):
    """
    Function to generate n random draws from a log normal distribution

    Args:
        n: integer number of draws to make
        med: median of distribrition
        s: standard deviation of distribution

    Returns:
        Vector of draws
    """
    sig = np.sqrt(np.log(1 + (s ** 2 / med ** 2)))
    draws = np.exp(np.log(med) + sig * np.random.randn(n))
    return draws


def normDraws(n, m, s):
    """
    Function to generate n random draws from a normal distribution

    Args:
        n: integer number of draws to make
        m: mean of distribrition
        s: standard deviation of distribution

    Returns:
        Vector of draws
    """
    draws = np.random.normal(m, s, n)
    return draws


def ensemble_pars_gen(n):
    """
    Function to generate perturbabion parameters to create an ensemble of meteorological forcings

    Args:
        n: size of the ensemble (integer)

    Returns:
        array of perturbation factors of dimension n
    """

    pbias = lognormDraws_kris(n, 1, 0.5)
    tbias = normDraws(n, 1, 0.01)  # this is for K only, we convert C to K then multiply bias then K to C (roughly +/- 4 degrees)
    swbias = normDraws(n, 1, 0.2)
    lwbias = normDraws(n, 1, 0.2)

    # uncorrelated
    perturb_uncor = pd.DataFrame({'pbias': pbias, 'tbias': tbias, 'swbias': swbias, 'lwbias': lwbias})

    cordat = np.array([1, -0.1, 0.5, -0.1, -0.1, 1, -0.3, 0.3, 0.5, -0.3, 1, 0.6, -0.1, 0.3, 0.6,
                       1])  # from Navari 2016 (not bad when compared to data), auto gen from all meteo in base run.
    # https://stats.stackexchange.com/questions/82261/generating-correlated-distributions-with-a-certain-mean-and-standard-deviation
    cormat = cordat.reshape(4, 4)

    # P,sw,lw,ta X P,sw,lw,ta  as table 1 navari 2016 https://tc.copernicus.org/articles/10/103/2016/tc-10-103-2016.pdf
    # array([[1., -0.1, 0.5, -0.1],
    #        [-0.1, 1., -0.3, 0.3],
    #        [0.5, -0.3, 1., 0.6],
    #        [-0.1, 0.3, 0.6, 1.]])

    # precip vals (pos 2) are the expected means sd of the lognorm distribution. we compute the mead sd of norm distribution below and substitute.
    sd_vec = [0.2, 0.2, 0.1, 0.005]  # p, sw, lw, ta Navari, how to base this on the data?
    mean_vec = np.array([0, 1, 1, 1])  # p, sw, lw, ta

    mean = mean_vec
    # empirical = true gives exact correlations in result eg https://stats.stackexchange.com/questions/82261/generating-correlated-distributions-with-a-certain-mean-and-standard-deviation
    # res = mvrnorm(n=N, m=mean_vec, Sigma=covmat, empirical=TRUE)

    # check for posite definiteness of covariance matrix
    # stopifnot(eigen(covmat)$values > 0 )

    # convert cor2cov matrix: https://stackoverflow.com/questions/39843162/conversion-between-covariance-matrix-and-correlation-matrix/39843276#39843276
    covmat = cormat * sd_vec

    cov = covmat  # diagonal covariance
    # from here https://stackoverflow.com/questions/31666270/python-analog-of-mvrnorms-empirical-setting-when-generating-a-multivariate-dist?rq=1
    res = np.random.multivariate_normal(mean, cov, n).T

    # lognorm transform on P
    plogN = np.exp(res[0,])
    tbias = res[3]  # this is for K only, we convert C to K then multiply bias then K to C
    pbias = plogN
    swbias = res[1,]
    lwbias = res[2,]
    perturb = pd.DataFrame({'pbias': pbias, 'tbias': tbias, 'swbias': swbias, 'lwbias': lwbias})
    perturb.to_csv("ensemble.csv")
    # df.plot(kind="density", subplots=True)
    # check distributions of random draws an correlations between variables
    pd.plotting.scatter_matrix(perturb, alpha=0.2)
    plt.show()
    return perturb, perturb_uncor


def ensemble_meteo_gen(ds, perturb, N_ens, ensemb_type):
    """
    Function to perturb downscaled meteorological timeseries

    Args:
        ds: downscaled_pts
        perturb: matrix of perturbation parameters
        N_ens: ensemble number (integer)
        ensemb_type: string describing which meteo variables to perturb: "T" = Tair, "P" = Precip, "S" = Swin, "L" = Lwin

    Returns:
        ds:

    """

    # create independent copy of ds to export
    ds_perturb = copy.deepcopy(ds)
    if ensemb_type == "T":
        ds_perturb["t"] = ds.t * perturb.tbias[N_ens]
    if ensemb_type == "TP":
        ds_perturb["t"] = ds.t * perturb.tbias[N_ens]
        ds_perturb["tp"] = ds.tp * perturb.pbias[N_ens]
    if ensemb_type == "TPS":
        ds_perturb["t"] = ds.t * perturb.tbias[N_ens]
        ds_perturb["tp"] = ds.tp * perturb.pbias[N_ens]
        ds_perturb["SW"] = ds.SW * perturb.swbias[N_ens]
    if ensemb_type == "TPSL":
        ds_perturb["t"] = ds.t * perturb.tbias[N_ens]
        ds_perturb["tp"] = ds.tp * perturb.pbias[N_ens]
        ds_perturb["SW"] = ds.SW * perturb.swbias[N_ens]
        ds_perturb["LW"] = ds.LW * perturb.lwbias[N_ens]

    return ds_perturb


def construct_HX(wdir, startDA, endDA):
    """
    Function to generate HX matrix of ensembleSamples x day

    Args:
        wdir: working directory
        DAyear: year of data assimilation yyyy
        startDA: start of data assimilation window  yyyy-mm-dd
        endDA: end of data assimilation window yyyy-mm-dd

    Returns:
        ensembleRes: 2D pandas dataframe with dims cols (samples * ensembles) and rows (days / simulation timesteps)

    """
    a = glob.glob(wdir + "/fsm_sims/sim_ENS*")

    # Natural sort to get correct order ensemb * samples
    def natural_sort(l):
        def convert(text): return int(text) if text.isdigit() else text.lower()
        def alphanum_key(key): return [convert(c) for c in re.split('([0-9]+)', key)]
        return sorted(l, key=alphanum_key)
    file_list = natural_sort(a)

    # compute dates index for da we always extract 31 March to Jun 30 for DA -
    # this may of course not always be appropriate
    df = pd.read_csv(file_list[0], delim_whitespace=True, parse_dates=[[0, 1, 2]], header=None)
    startIndex = df[df.iloc[:, 0] == startDA].index.values
    endIndex = df[df.iloc[:, 0] == endDA].index.values

    # if len(startIndex) == 0:
    #     sys.exit("daYear does not exist in simulation period")
    # if len(endIndex) == 0:
    #     sys.exit("daYear does not exist in simulation period")

    # construct a list as long as samples x ensembles (sampleEnsembles)
    data = []
    for file_path in file_list:
        # column 7 (index 6) is snow water equiv
        data.append(np.genfromtxt(file_path, usecols=6)[int(startIndex):int(endIndex) + 1])

    myarray = np.asarray(data)  # sampleEnsembles x days
    ensembRes = pd.DataFrame(myarray.transpose())

    mydates = df.iloc[:, 0][int(startIndex):int(endIndex) + 1]

    return ensembRes, mydates

    # data = []
    # for file_path in file_list:
    #     # column 6 (index 5) is snow height
    #     data.append(np.genfromtxt(file_path, usecols=5)[int(startIndex):int(endIndex)])
    #
    # myarray = np.asarray(data)
    # df = pd.DataFrame(myarray.transpose())
    #
    # df.to_csv(wd + '/ensemble/ensembResHS' + str(ipad) + '.csv', index=False, header=False)
    #
    # # get open - loop results, but only once when ensembleN=0
    # if int(ensembleN) == 1:
    #
    #     a = glob.glob(wd + "/fsm_sims/*")
    #     file_listOL = natural_sort(a)
    #
    #     df = pd.read_csv(file_listOL[0], delim_whitespace=True,
    #                      parse_dates=[[0, 1, 2]], header=None)
    #     startIndex = df[df.iloc[:, 0] == str(daYear - 1) + "-09-01"].index.values
    #     endIndex = df[df.iloc[:, 0] == str(daYear) + "-09-01"].index.values
    #     time = df.iloc[int(startIndex):int(endIndex), 0]
    #
    #     data = []
    #     for file_path in file_listOL:
    #         data.append(np.genfromtxt(file_path, usecols=6)[int(startIndex):int(endIndex)])
    #
    #     myarray = np.asarray(data)
    #     df2 = pd.DataFrame(myarray.transpose())
    #
    #     # 2d rep of 3d ensemble in batches where ensemble changes slowest
    #     df2.to_csv(wd + '/ensemble/openloopRes.csv', index=False, header=False)
    #     time.to_csv(wd + '/ensemble/ensemble_timestamps.csv', index=False, header=False)


def modis_tile(latmax, latmin, lonmax, lonmin):
    """
    Function to retrieve MODIS tile indexes based on a lon/lat bbox

    Args:


    Returns:


    """

    # find modis tiles based on longlat

    os.system("wget --no-check-certificate https://modis-land.gsfc.nasa.gov/pdf/sn_bound_10deg.txt")

    # first seven rows contain header information
    # bottom 3 rows are not data
    data = np.genfromtxt('sn_bound_10deg.txt',
                         skip_header=7,
                         skip_footer=3)
    in_tile = False
    i = 0
    while (not in_tile):
        in_tile = latmax >= data[i, 4] and latmax <= data[i, 5] and lonmax >= data[i, 2] and lonmax <= data[i, 3]
        i += 1
    vertmax = data[i - 1, 0]
    horizmax = data[i - 1, 1]
    print('max Vertical Tile:', vertmax, 'max Horizontal Tile:', horizmax)

    in_tile = False
    i = 0
    while (not in_tile):
        in_tile = latmin >= data[i, 4] and latmin <= data[i, 5] and lonmin >= data[i, 2] and lonmin <= data[i, 3]
        i += 1
    vertmin = data[i - 1, 0]
    horizmin = data[i - 1, 1]
    print('min Vertical Tile:', vertmin, 'min Horizontal Tile:', horizmin)

    return vertmax, horizmax, vertmin, horizmin


def get_modis_wrapper(vertmax, horizmax, vertmin, horizmin, startDate, endDate):
    """
    Function to iterate through several MODIS tile downloads.

    Args:


    Returns:


    """


def pymodis_download(wdir, vert, horiz, STARTDATE, ENDDATE):
    # http://www.pymodis.org/scripts/modis_download.html

    USER = "jfiddes"  # os.getenv('user')
    PWD = "sT0kkang!"  # os.getenv('pwd')
    MODPATH = wdir + "/modis/"
    HOST = "https://n5eil01u.ecs.nsidc.org/"
    FOLDER = "MOST"
    PRODUCT = "MOD10A1F.061"
    RAWPATH = MODPATH + "/raw/"
    os.makedirs(MODPATH, exist_ok=True)
    os.makedirs(RAWPATH, exist_ok=True)
    os.makedirs(MODPATH + "/transformed/", exist_ok=True)

    hreg = ("h" + f"{horiz:02d}")  # re.compile("h2[4]")
    vreg = ("v" + f"{vert:02d}")  # re.compile("v0[5]")
    TILE = hreg + vreg
    print("Downloading " + product + " from " + STARTDATE + " to " + ENDDATE)
    os.system("modis_download.py -U " +
              USER +
              " -P " + PWD +
              " -u " + HOST +
              " -p mail@pymodis.com "
              " -s " + FOLDER +
              " -p " + PRODUCT +
              " -f " + STARTDATE +
              " -e " + ENDDATE +
              " -t " + TILE + " " +
              RAWPATH)


def projFromLandform(pathtolandform):
    src = gdal.Open(pathtolandform, gdal.GA_ReadOnly)
    proj = osr.SpatialReference(wkt=src.GetProjection())
    epsg = proj.GetAttrValue('AUTHORITY', 1)

    ulx, xres, xskew, uly, yskew, yres = src.GetGeoTransform()
    lrx = ulx + (src.RasterXSize * xres)
    lry = uly + (src.RasterYSize * yres)
    bbox = [ulx, lry, lrx, uly]

    return epsg, bbox


def process_modis(wdir, epsg, bbox, layer=0):
    """
    Function to convert hdf to tiff, extract layer, and reproject

    Args:
        modpath: path to 'modis' directory
        layer: layer to extract from hdf
        epsg: reprojected target epsg code
        bbox: output bounds as list [minX, minY, maxX, maxY] in target

    Returns:
        Spatially subsetted reprojected tiffs of data layer specified (still NDSI)


        Docs: https://nsidc.org/sites/nsidc.org/files/technical-references/C6.1_MODIS_Snow_User_Guide.pdf
    """

    MODPATH = wdir + "/modis/"
    RAWPATH = MODPATH + "/raw/"

    # list all downloaded hdf files
    files = glob.glob(RAWPATH + "/*.hdf")

    # extract
    for in_file in files:
        out_file = MODPATH + "/transformed/" + os.path.basename(in_file).replace('.hdf', '.tif')

        # open dataset
        dataset = gdal.Open(in_file, gdal.GA_ReadOnly)

        # extract subset layer [default = 0, this is The cloud-gap-filled (CGF) daily snow cover tiled product.]
        subdataset = gdal.Open(dataset.GetSubDatasets()[layer][0], gdal.GA_ReadOnly)

        # gdalwarp to longlat
        # kwargs = {'format': 'GTiff', 'ouputBounds': str(bbox), 'dstSRS': 'EPSG:'+ str(epsg)}
        # ds = gdal.Warp(destNameOrDestDS=out_file, srcDSOrSrcDSTab=subdataset, **kwargs)
        ds = gdal.Warp(destNameOrDestDS=out_file,
                       srcDSOrSrcDSTab=subdataset,
                       format='GTiff',
                       outputBounds=bbox,
                       outputBoundsSRS='EPSG:' + str(epsg),
                       dstSRS='EPSG:' + str(epsg))
        del ds

    return print(str(len(files)) + " MODIS files processed")


def extract_fsca_timeseries(wdir, plot=True):
    """
    Function to extract mean NDSI from domain and converts to fSCA using the empirical formula based on landsat
    https://modis-snow-ice.gsfc.nasa.gov/uploads/C6_MODIS_Snow_User_Guide.pdf pp.22
    Args:
        wdir: working directory
        plot: booleen whether to plot dataframe

    Returns:
        df_sort: dataframe of julian dates and fSCA values
    """

    files = glob.glob(wdir + "/modis/transformed/*.tif")
    data = []
    juldates = []
    for file in files:
        with rio.open(file) as src:
            # Read as numpy array
            array = src.read()
            profile = src.profile

            mymean = np.mean(array)
            juldate = file.split(".")[1].split("A")[1]

        data.append(mymean)
        juldates.append(juldate)

    # convert to fSCA
    fsca = (-0.01 + (1.45 * np.array(data)))
    fsca[fsca <= 0] = 0
    fsca[fsca > 100] = 100

    d = {'dates': juldates, 'fSCA': fsca}
    df = pd.DataFrame(data=d)
    df_sort = df.sort_values("dates")
    # add datetime object
    df_sort["mydates"] = [pd.to_datetime(e[:4]) + pd.to_timedelta(int(e[4:]) - 1, unit='D') for e in df_sort.dates]

    df_sort_reindex = df_sort.reset_index(drop=True)
    if plot:
        df_sort_reindex.plot(x="mydates", y="fSCA", kind="scatter")
        plt.show()

    return df_sort_reindex

    # import datetime
    # datetime.datetime.strptime(df_sort.dates, '%Y%j').date()
    # df_sort["mydates"]


# def get_modis(vert, horiz, startYear, endYear):
#     """
#     Function to retrieve MODIS tile indexes based on a lon/lat bbox
#
#     Args:
#
#
#     Returns:
#
#
#     """
#
#     # require MODIS downloaded
#     # https://bpostance.github.io/posts/processing-large-spatial-datasets.md/
#
#     # format h,v
#
#     import os
#     import re
#     import time
#     import numpy as np
#     import rasterio as rio
#     from osgeo import gdal
#     import requests
#     from bs4 import BeautifulSoup # pip install beautifulsoup4
#     import dotenv # pip install python-dotenv
#     dotenv.load_dotenv(dotenv.find_dotenv())
#
#     # params fo LP DAAC
#     # you will require an EarthData login
#     host = fr'https://e4ftl01.cr.usgs.gov/MOTA/MCD64A1.061'
#     host = fr'https://n5eil01u.ecs.nsidc.org/MOST/MOD10A1F.061/'
#     login = "jfiddes" # os.getenv('user')
#     password = "sT0kkang!" # os.getenv('pwd')
#     modpath = "/home/joel/sim/topoPyscale_davos/modis/"
#     os.makedirs(modpath, exist_ok=True)
#     os.makedirs(modpath + "/raw/", exist_ok=True)
#     os.makedirs(modpath + "/transformed/", exist_ok=True)
#
#     # list folders
#     r = requests.get(host, verify=True, stream=True,auth=(login,password))
#     soup = BeautifulSoup(r.text, "html.parser")
#     folders = list()
#     for link in soup.findAll('a', attrs={'href': re.compile("\d{4}.\d{2}.\d{2}/")}):
#         folders.append(link.get('href'))
#     print(f"{len(folders)} folders found")
#
#     # subset folders by date
#     startDate = str(startYear) + ".09.01/"
#     endDate = str(endYear) + ".01.31/"
#
#     s = folders.index(startDate)
#     e = folders.index(endDate)
#
#
#     # list files in folder
#     from timeit import default_timer as timer
#     from datetime import timedelta
#
#     start = timer()
#
#     # 30 sec per 10 days
#     for f in folders[1:5]:
#         file_list = list()
#         folder_url = f"{host}/{f}"
#         r = requests.get(folder_url, verify=True, stream=True, auth=(login,password))
#         soup = BeautifulSoup(r.text, "html.parser")
#         for link in soup.findAll('a', attrs={'href': re.compile(".hdf$")}):
#             file_list.append(link.get('href'))
#     print(f"{len(file_list)} files found in folder n")
#
#     end = timer()
#     print(timedelta(seconds = end - start))
#
#     # USA tiles only
#     # use modis grid to slice only continents / areas of interest
#     # https://modis-land.gsfc.nasa.gov/MODLAND_grid.html
#     hreg = re.compile("h" + f"{horiz:02d}") # re.compile("h2[4]")
#     vreg = re.compile("v" + f"{vert:02d}" )# re.compile("v0[5]")
#     usa_files = list()
#     for fl in file_list:
#         h = hreg.search(fl)
#         if h:
#             v = vreg.search(h.string)
#             if v:
#                 usa_files.append(v.string)
#     print(f"{len(usa_files)} USA tiles found")
#
#     # download file from folder
#     # implement 25 file timeout to avoid rate limit
#     exceptions = list()
#     for f in folders[:1]:
#         for e, fl in enumerate(usa_files[:]):
#             if (e+1) % 10 == 0:
#                 print('sleep')
#                 time.sleep(5)
#             else:
#                 print(fl, e+1)
#                 try:
#                     file_url = f"{host}/{f}/{fl}"
#                     r = requests.get(file_url, verify=True, stream=True,auth=(login,password))
#                     open(modpath + "/raw/" + fl, "wb").write(r.content)
#                 except Exception as error:
#                     print(error)
#                     exceptions.append(fl)
#
#     for fl in usa_files[:]:
#         in_file = modpath + "raw/" + fl
#         out_file = modpath + "/transformed/" + fl.replace('.hdf', '.tif')
#
#         # open dataset
#         dataset = gdal.Open(in_file, gdal.GA_ReadOnly)
#         subdataset = gdal.Open(dataset.GetSubDatasets()[0][0], gdal.GA_ReadOnly)
#
#         # gdalwarp
#         kwargs = {'format': 'GTiff', 'dstSRS': 'EPSG:4326'}
#         ds = gdal.Warp(destNameOrDestDS=out_file, srcDSOrSrcDSTab=subdataset, **kwargs)
#         del ds

def particle_batch_smoother(Y, HX, R):
    # particle batch smoothert
    No = len(Y)
    Rinv = np.array([(1 / R)] * No)
    mat = np.ones((HX.shape[0], HX.shape[1]))

    # reshape Y to have dims of mat
    Y1 = np.reshape(np.array(Y), (HX.shape[0], 1))

    # Calculate the likelihood.
    Inn = np.kron(mat, Y1) - HX

    # with %*% gives NA
    EObj = Rinv.dot(Inn ** 2)  # [1 x Ne] ensemble objective function.

    LH = np.exp(-0.5 * EObj)  # Scaled likelihood.

    # Calculate the posterior weights as the normalized likelihood.
    w = LH / sum(LH)


import numpy as np
import matplotlib.pyplot as plt
from scipy.special import logsumexp


def EnKA(prior, obs, pred, alpha, R):
    """
    EnKA: Implmentation of the Ensemble Kalman Analysis
    Inputs:
        prior: Prior ensemble matrix (n x N array)
        obs: Observation vector (m x 1 array)
        pred: Predicted observation ensemble matrix (m x N array)
        alpha: Observation error inflation coefficient (scalar)
        R: Observation error covariance matrix (m x m array, m x 1 array, or scalar)
    Outputs:
        post: Posterior ensemble matrix (n x N array)
    Dimensions:
        N is the number of ensemble members, n is the number of state variables and/or
        parameters, and m is the number of boservations.

    For now, we have impelemnted the classic (stochastic) version of the Ensemble
    Kalman analysis that involves perturbing the observations. Deterministic variants
    (Sakov and Oke, 2008) also exist that may be worth looking into. This analysis
    step can be used both for filtering, i.e. Ensemble Kalman filter, when observations
    are assimilated sequentially or smoothing, i.e. Ensemble (Kalman) smoother,
    when also be carried out in a multiple data assimilation (i.e. iterative) approach.

    Since the author rusty in python it is possible that this code can be improved
    considerably. Below are also some examples of further developments that may be
    worth looking into:

    To do/try list:
    (i) Using a Bessel correction (Ne-1 normalization) for sample covariances.
    (ii) Perturbing the predicted observations rather than the observations
        following van Leeuwen (2020, doi: 10.1002/qj.3819) which is the formally
        when defining the covariances matrices.
    (iii) Looking into Sect. 4.5.3. in Evensen (2019, doi: 10.1007/s10596-019-9819-z)
    (iv) Using a deterministic rather than stochastic analysis step.
    (v) Augmenting with artificial momentum to get some inertia.
    (vi) Better understand/build on the underlying tempering/annealing idea in ES-MDA.
    (vii) Versions that may scale better for large n such as those working in the
        ensemble subspace (see e.g. formulations in Evensen, 2003, Ocean Dynamics).

    Code by: K. Aalstad (10.12.2020) based on an earlier Matlab version.
    """

    # Dimensions.
    m = np.size(obs)  # Number of obs
    n = np.shape(prior)[0]  # Tentative number of state vars and/or parameters
    if np.size(prior) == n:  # If prior is 1D correct the above.
        N = n  # Number of ens. members.
        n = 1  # Number of state vars and/or parameters.
    else:  # If prior is 2D then the lenght of the second dim is the ensemble size
        N = np.shape(prior)[1]

    # Checks on the observation error covariance matrix.
    try:
        if np.shape(R)[0] == m and np.shape(R)[1] == m:
            pass
            # Do nothing if specified as a matrix.
    except:
        if np.size(R) == 1:
            R = R * np.identity(m)
        elif np.size(R) == m:
            R = np.diag(R)
            # Convert to matrix if specified as a vector.
        else:
            raise Exception('R must be a scalar, m x 1 vector, or m x m matrix.')

    # pdb.set_trace()
    # Anomaly calculations.
    mprior = np.mean(prior, -1)  # Prior ensemble mean
    if n == 1:
        A = prior - mprior
    else:
        A = prior - mprior[:, None]
    mpred = np.mean(pred, -1)  # Prior predicted obs ensemble mean
    if m == 1:
        B = pred - mpred
    else:
        B = pred - mpred[:, None]
    Bt = B.T  # Tranposed -"-

    # Perturbed observations.
    Y = np.outer(obs, np.ones(N)) + np.sqrt(alpha) * np.random.randn(m, N)
    # Y=np.outer(obs,np.ones(N))

    # Covariance matrices.
    C_AB = A @ Bt  # Prior-predicted obs covariance matrix multiplied by N (n x m)
    C_BB = B @ Bt  # Predicted obs covariance matrix multiplied by N (m x m)
    aR = (N * alpha) * R  # Scaled observation error covariance matrix (m x m)

    # Analysis step
    if n == 1 and m == 1:  # Scalar case
        K = C_AB * (np.linalg.inv(C_BB + aR))
        inno = Y - pred
        post = prior + K * inno
    else:
        K = C_AB @ (np.linalg.inv(C_BB + aR))  # Kalman gain (n x m)
        inno = Y - pred  # Innovation (m x N)
        post = prior + K @ inno  # Posterior (n x N)

    return post


def PBS(obs, pred, R):
    """
    PBS: Implmentation of the Particle Batch Smoother
    Inputs:
        obs: Observation vector (m x 1 array)
        pred: Predicted observation ensemble matrix (m x N array)
        R: Observation error covariance 'matrix' (m x 1 array, or scalar)
    Outputs:
        w: Posterior weights (N x 1 array)
    Dimensions:
        N is the number of ensemble members and m is the number of boservations.

    Here we have implemented the particle batch smoother, which is a batch-smoother
    version of the particle filter (i.e. a particle filter without resampling),
    described in Margulis et al. (2015, doi: 10.1175/JHM-D-14-0177.1). As such, this
    routine can also be used for particle filtering with sequential data assimilation.
    This scheme is obtained by using a particle (mixture of Dirac delta functions)
    representation of the prior and applying this directly to Bayes theorem. In other
    words, it is just an application of importance sampling with the prior as the
    proposal density (importance distribution). It is also the same as the Generalized
    Likelihood Uncertainty Estimation (GLUE) technique (with a formal Gaussian likelihood)
    which is widely used in hydrology.

    This particular scheme assumes that the observation errors are additive Gaussian
    white noise (uncorrelated in time and space). The "logsumexp" trick is used to
    deal with potential numerical issues with floating point operations that occur
    when dealing with ratios of likelihoods that vary by many orders of magnitude.

    To do/try list:
        (i) Implement an option for correlated observation errors if R is specified
        as a matrix (this would slow down this function).
        (ii) Consier other likelihoods and observation models.
        (iii) Look into iterative versions of importance sampling.


    Code by: K. Aalstad (14.12.2020) based on an earlier Matlab version.
    """
    # Dimensions.
    m = np.size(obs)  # Number of obs
    N = np.shape(pred)[-1]

    # Checks on the observation error covariance matrix.
    if np.size(R) == 1:
        R = R * np.ones(m)
    elif np.size(R) == m:
        pass
    else:
        raise Exception('R must be a scalar, m x 1 vector.')

    # pdb.set_trace()
    # Residual and log-likelihood
    if m == 1:
        residual = obs - pred
        llh = -0.5 * ((residual ** 2) * (1 / R))
    else:
        residual = obs[:, None] - pred
        # pdb.set_trace()
        llh = -0.5 * ((1 / R) @ (residual ** 2))

    # Log of normalizing constant
    # A properly scaled version of this could be output for model comparison.
    lZ = logsumexp(llh)  # from scipy.special import logsumexp

    # Weights
    logw = llh - lZ  # Log of posterior weights
    w = np.exp(logw)  # Posterior weights
    if np.size(w) == N and np.round(np.sum(w), 10) == 1:
        pass
    else:
        raise Exception('Something went wrong with the PBS, try pdb.set_trace()')

    return w


def da_plots(HX1, HX2, W, mydates, myobs):
    """
    Function to extract mean fSSCA from domain

    Args:
        HX1:
        HX2:
        weights: weights output from PBS (dims 1 x NENS)
        dates: datetime object
        myobs: timeseries of observtions

    Returns:
        plot
    """

    # HX1
    #
    # weights = W
    # HX = HX3D_weighted_mean

    def percentile_plot(prior, wfill, percentile):
        ndays = prior.shape[0]
        med_post = []
        for days in range(ndays):
            mu = prior[days, ]
            d = {'mu': mu, 'wfill': wfill}
            df = pd.DataFrame(d)
            dfOrder = df.sort_values('mu')
            y_interp = interp1d(np.cumsum(dfOrder.wfill), dfOrder.mu, fill_value="extrapolate")
            med = y_interp(percentile)
            med_post.append(med)

        return np.array(med_post)

    # for prior all ensemble members considered equally likely therefore weights are all
    weights = [1/HX1.shape[1]] * HX1.shape[1]
    prior_med = percentile_plot(HX1, weights, 0.5)
    prior_low = percentile_plot(HX1, weights, 0.1)
    prior_high = percentile_plot(HX1, weights, 0.8)

    # for posterior weights are given by the output of PBS, W
    post_med = percentile_plot(HX1, W, 0.5)
    post_low = percentile_plot(HX1, W, 0.1)
    post_high = percentile_plot(HX1, W, 0.8)

    plt.subplot(1, 2, 1)
    plt.fill_between(mydates, prior_low, prior_high, alpha=0.2)
    plt.plot(mydates, prior_med)
    plt.fill_between(mydates, post_low, post_high, alpha=0.2)
    plt.plot(mydates, post_med)
    plt.plot(mydates, myobs)
    plt.xlabel("Date")
    plt.ylabel("fSCA")
    plt.legend()

    # for prior all ensemble members considered equally likely therefore weights are all
    weights = [1/HX1.shape[1]] *  HX1.shape[1]
    prior_med = percentile_plot(HX2, weights, 0.5)
    prior_low= percentile_plot(HX2, weights, 0.1)
    prior_high = percentile_plot(HX2, weights, 0.9)

    # for posterior weights are given by the output of PBS, W
    post_med = percentile_plot(HX2, W, 0.5)
    post_low = percentile_plot(HX2, W, 0.1)
    post_high = percentile_plot(HX2, W, 0.9)

    plt.subplot(1, 2, 2)
    plt.fill_between(mydates, prior_low, prior_high, alpha=0.2)
    plt.plot(mydates, prior_med)
    plt.fill_between(mydates, post_low, post_high, alpha=0.2)
    plt.plot(mydates, post_med)
    plt.xlabel("Date")
    plt.ylabel("HS (m)")
    plt.legend()
    plt.show()


def fsca_plots(wdir, plotDay, df_mean):
    """
    Function to plot side by side observed fSCA 500m [0-1] and modelled SCA 30 m [0,1]

    Args:
        wdir: sim directory
        plotDay: date to plot as YYYY-mm-dd
        df_mean: timeseries mean object given by sim.timeseries_means_period()

    Returns:
        plot

    Usage:
        df = sim.agg_by_var_fsm(6)
        df2 = sim.agg_by_var_fsm_ensemble(6, W)
        df_mean = sim.timeseries_means_period(df, "2018-06-01", "2018-06-01")
        da.fsca_plots("/home/joel/sim/topoPyscale_davos/", "2018150", df_mean)

    TODO:
        - create modelled fSCA option from agreggated 30m to 500m pixels
        - accept datetime object instead of julday.
    """
    # convert data YYYY-mm-dd to YYYYddd
    dt = datetime.strptime(plotDay, "%Y-%m-%d")
    doy = dt.timetuple().tm_yday
    year = dt.year
    juldate = str(year)+str(doy)


    modis_file = glob.glob(wdir + "modis/transformed/MOD10A1F.A" + juldate + "*")

    src = rio.open(modis_file[0])
    fig, (axrgb, axhist) = plt.subplots(1, 2, figsize=(14,7))

    rio.plot.show(src, ax=axrgb,vmin=0, vmax=100)
    cbar1 = plt.imshow(src.read(1), cmap='viridis', vmin=0, vmax=100)
    fig.colorbar(cbar1, ax=axrgb)

    nclust = len(df_mean)
    lookup = np.arange(nclust, dtype=np.uint16)

    for i in range(0, nclust):
        lookup[i] = df_mean[i]

    with rasterio.open('landform.tif') as src:
        # Read as numpy array
        array = src.read()
        profile = src.profile

        # Reclassify in a single operation using broadcasting
        array = lookup[array]
        # binary map
        array[array <= 0] = 0
        array[array > 0] = 1


    # rasterio.plot.show(array, cmap='viridis')
    # plt.show()

    with rio.open('output_raster.tif', 'w', **profile) as dst:
        # Write to disk
        dst.write(array.astype(rasterio.int16))

    src1 = rio.open("output_raster.tif")
    rio.plot.show(src1, ax=axhist)
    cbar2 = plt.imshow(src1.read(1), cmap='viridis')
    fig.colorbar(cbar2, ax=axhist)
    plt.show()


# def spatial_plots(df_mean, plot=False):
#     """
#     Function to plot side by side observed fSCA 500m [0-1] and modelled SCA 30 m [0,1]
#
#     Args:
#         df_mean: timeseries mean object given by sim.timeseries_means_period()
#         plot: genrate plot? logical
#
#     Returns:
#         plot
#
#     Usage:
#         df = sim.agg_by_var_fsm(6)
#         df2 = sim.agg_by_var_fsm_ensemble(6, W)
#         df_mean = sim.timeseries_means_period(df, "2018-06-01", "2018-06-01")
#         da.fsca_plots("/home/joel/sim/topoPyscale_davos/", "2018150", df_mean)
#
#     TODO:
#         - create modelled fSCA option from agreggated 30m to 500m pixels
#     """
#
#
#     nclust = len(df_mean)        wdir: sim directory
#         plotDay: date to plot as YYYY-mm-dd
#     lookup = np.arange(nclust, dtype=np.uint16)
#
#     for i in range(0, nclust):
#         lookup[i] = df_mean[i]
#
#     with rasterio.open('landform.tif') as src:
#         # Read as numpy array
#         array = src.read()
#         profile = src.profile
#
#         # Reclassify in a single operation using broadcasting
#         array = lookup[array]
#         # binary map
#         array[array <= 0] = 0
#         array[array > 0] = 1
#
#
#     # rasterio.plot.show(array, cmap='viridis')
#     # plt.show()
#
#     with rio.open('output_raster.tif', 'w', **profile) as dst:
#         # Write to disk
#         dst.write(array.astype(rasterio.int16))
#
#     if plot:
#         src1 = rio.open("output_raster.tif")
#         rio.plot.show(src1, ax=axhist)
#         cbar2 = plt.imshow(src1.read(1), cmap='viridis')
#         fig.colorbar(cbar2, ax=axhist)
#         plt.show()

def da_compare_plot(wdir, plotDay, df_mean_open, df_mean_da):
    """
    Function to plot side by side observed fSCA 500m [0-1] and modelled SCA 30 m [0,1]

    Args:
        wdir: sim directory
        plotDay: date to plot as YYYY-mm-dd
        df_mean: timeseries mean object given by sim.timeseries_means_period()

    Returns:
        plot

    Usage:
        df = sim.agg_by_var_fsm(6)
        df2 = sim.agg_by_var_fsm_ensemble(6, W)
        df_mean = sim.timeseries_means_period(df, "2018-06-01", "2018-06-01")
        da.fsca_plots("/home/joel/sim/topoPyscale_davos/", "2018150", df_mean)

    TODO:
        - create modelled fSCA option from agreggated 30m to 500m pixels
        - accept datetime object instead of julday.
    """
    # set up plot axis
    fig, ([axmod1, axopen], [axmod2, axda]) = plt.subplots(2, 2, figsize=(14, 7))

    # convert data YYYY-mm-dd to YYYYddd
    dt = datetime.strptime(plotDay, "%Y-%m-%d")
    doy = dt.timetuple().tm_yday
    year = dt.year
    juldate = str(year)+str(doy)

    # plot modis
    modis_file = glob.glob(wdir + "modis/transformed/MOD10A1F.A" + juldate + "*")
    src = rio.open(modis_file[0])
    rio.plot.show(src, ax=axmod1,vmin=0, vmax=100)
    cbar1 = plt.imshow(src.read(1), cmap='viridis', vmin=0, vmax=100)
    fig.colorbar(cbar1, ax=axmod1)

    # compute and plot open loop
    nclust = len(df_mean_open)
    lookup = np.arange(nclust, dtype=np.uint16)

    for i in range(0, nclust):
        lookup[i] = df_mean_open[i]

    with rasterio.open('landform.tif') as src:
        # Read as numpy array
        array = src.read()
        profile = src.profile

        # Reclassify in a single operation using broadcasting
        array = lookup[array]
        # binary map
        #array[array <= 0] = 0
        #array[array > 0] = 1


    # rasterio.plot.show(array, cmap='viridis')
    # plt.show()

    with rio.open('output_raster.tif', 'w', **profile) as dst:
        # Write to disk
        dst.write(array.astype(rasterio.int16))

    src1 = rio.open("output_raster.tif")
    rio.plot.show(src1, ax=axopen)
    cbar2 = plt.imshow(src1.read(1), cmap='viridis')
    fig.colorbar(cbar2, ax=axopen)

    # plot modis again
    src = rio.open(modis_file[0])
    rio.plot.show(src, ax=axmod2,vmin=0, vmax=100)
    cbar1 = plt.imshow(src.read(1), cmap='viridis', vmin=0, vmax=100)
    fig.colorbar(cbar1, ax=axmod2)

    # compute and plot DA
    nclust = len(df_mean_da)
    lookup = np.arange(nclust, dtype=np.uint16)

    for i in range(0, nclust):
        lookup[i] = df_mean_da[i]

    with rasterio.open('landform.tif') as src:
        # Read as numpy array
        array = src.read()
        profile = src.profile

        # Reclassify in a single operation using broadcasting
        array = lookup[array]
        # binary map
        #array[array <= 0] = 0
        #array[array > 0] = 1


    # rasterio.plot.show(array, cmap='viridis')
    # plt.show()

    with rio.open('output_raster.tif', 'w', **profile) as dst:
        # Write to disk
        dst.write(array.astype(rasterio.int16))

    src1 = rio.open("output_raster.tif")
    rio.plot.show(src1, ax=axda)
    cbar2 = plt.imshow(src1.read(1), cmap='viridis')
    fig.colorbar(cbar2, ax=axda)
    plt.show()

def build_modis_cube(wdir):
    geotiff_list_unsorted = glob.glob(wdir + "modis/transformed/MOD10A1F.A*")
    # Natural sort to get correct order ensemb * samples
    def natural_sort(l):
        def convert(text): return int(text) if text.isdigit() else text.lower()
        def alphanum_key(key): return [convert(c) for c in re.split('([0-9]+)', key)]
        return sorted(l, key=alphanum_key)
    geotiff_list = natural_sort(geotiff_list_unsorted)

    # Create variable used for time axis
    mydates = [lname.split('.A')[1].split(".h")[0] for lname in geotiff_list]
    time_var = xr.Variable('time', mydates)
    # Load in and concatenate all individual GeoTIFFs
    geotiffs_da = xr.concat([xr.open_rasterio(i) for i in geotiff_list], dim=time_var)
    # Covert our xarray.DataArray into a xarray.Dataset
    geotiffs_ds = geotiffs_da.to_dataset('band')
    # Rename the variable to a more useful name
    ds = geotiffs_ds.rename({1: 'fSCA'})
    arr = ds.to_array()

    # tuples of coordinates
    x, y = np.meshgrid(range(len(arr.x)), range(len(arr.y)))

    myobs = []
    for index in range(len(x.flatten() )):
        xindex = x.flatten()[index]
        yindex = y.flatten()[index]
        obs = arr[0, :, range(len(arr.y))[yindex] ,range(len(arr.x))[xindex]]
        myobs.append(obs.values)

    # obs array of Npixels x time
    obs_array = np.array(myobs).shape


def getModisbbox(Npixel, arr):
    """
    Function to find obtain bbox of modis pixel CRS is projected UTM

    Args:
        Npixel: pixel nuber as integer of 0 - (xdim x ydim) (the flattened grid) eg for a 10 x 10 grid Npixel would have a range of 0-100
        arr: xarray DataArray with dimensions "x" and "y" and attribute "res"

    Returns:
        bbox: bbox of pixel as list [xmax,xmin, ymax, ymin]



    """

    xres = arr.res[0] # resolution in x
    yres = arr.res[1] # resolution in y

    x, y = np.meshgrid(arr.x, arr.y)

    centreX = x.flatten()[Npixel] # pixel N centre x coord
    centreY = y.flatten()[Npixel] # pixel N centre y coord
    xmax = float(centreX + xres/2)
    xmin = float(centreX - xres/2)
    ymax = float(centreY + yres/2)
    ymin = float(centreY - yres/2)
    bbox = [xmax,xmin, ymax, ymin]
    return bbox

def getSamples_inModisPixel(map_path):
    """
    Identify samples that exist within a single bbox and proportianal cover. This can be used to construct a modelled fSCA in a MODIS pixel

    Args:
        map_path: path to landform.tif

    Returns:
        list of samples and frequency

    """

    # # define proj
    # proj = pyproj.Transformer.from_crs(int(epsg_in), int(epsg_out), always_xy=True)
    # # project
    # x2, y2 = proj.transform(x_coord, y_coord)
    # open map
    dataset = rasterio.open(map_path)
    # get pixel x+y of the coordinate
    py, px = dataset.index(x2, y2)
    # create 1x1px window of the pixel
    window = rasterio.windows.Window(px - 1 // 2, py - 1 // 2, 1, 1)
    # read rgb values of the window
    clip = dataset.read(window=window)
    return (int(clip))




"""
Set of functions to work with DEMs
S. Filhol, Oct 2021

TODO:
- an improvement could be to first copmute horizons, and then SVF to avoid computing horizon twice
"""

import sys
import rasterio
from pyproj import Transformer
import pandas as pd
import numpy as np
import xarray as xr
from topocalc import gradient
from topocalc import viewf
from topocalc import horizon
import time
from multiprocessing.dummy import Pool as ThreadPool
import multiprocessing as mproc
import os
from TopoPyScale import topo_export as te


def convert_epsg_pts(xs,ys, epsg_src=4326, epsg_tgt=3844):
    """
    Simple function to convert a list fo poitn from one projection to another oen using PyProj

    Args:
        xs (array): 1D array with X-coordinate expressed in the source EPSG
        ys (array): 1D array with Y-coordinate expressed in the source EPSG
        epsg_src (int): source projection EPSG code
        epsg_tgt (int): target projection EPSG code

    Returns: 
        array: Xs 1D arrays of the point coordinates expressed in the target projection
        array: Ys 1D arrays of the point coordinates expressed in the target projection
    """
    print('Convert coordinates from EPSG:{} to EPSG:{}'.format(epsg_src, epsg_tgt))
    trans = Transformer.from_crs("epsg:{}".format(epsg_src), "epsg:{}".format(epsg_tgt), always_xy=True)
    Xs, Ys = trans.transform(xs, ys)
    return Xs, Ys


def get_extent_latlon(dem_file, epsg_src):
    """
    Function to extract DEM extent in Lat/Lon

    Args:
        dem_file (str): path to DEM file (GeoTiFF)
        epsg_src (int): EPSG projection code

    Returns: 
        dict: extent in lat/lon, {latN, latS, lonW, lonE}
    """
    with rasterio.open(dem_file) as rf:
        xs, ys = [rf.bounds.left, rf.bounds.right], [rf.bounds.bottom, rf.bounds.top]
    trans = Transformer.from_crs("epsg:{}".format(epsg_src), "epsg:4326", always_xy=True)
    lons, lats = trans.transform(xs, ys)
    extent = {'latN': lats[1],
              'latS': lats[0],
              'lonW': lons[0],
              'lonE': lons[1]}
    return extent


def extract_pts_param(df_pts, ds_param, method='nearest'):
    """
    Function to sample DEM parameters for a list point. This is used as an alternative the the TopoSub method, to perform downscaling at selected locations (x,y)
    WARNING: the projection and coordiante system of the EDM and point coordinates MUST be the same!

    Args:
        df_pts (dataframe): list of points coordinates with coordiantes in (x,y).
        ds_param (dataset): dem parameters
        method (str): sampling method. Supported 'nearest', 'linear' interpolation, 'idw' interpolation (inverse-distance weighted)

    Returns: 
        dataframe: df_pts updated with new columns ['elevation', 'slope', 'aspect', 'aspect_cos', 'aspect_sin', 'svf']
    """
    print('\n---> Extracting DEM parameters for the given list of point coordinates')
    # delete columns in case they already exist
    df_pts = df_pts.drop(['elevation', 'slope', 'aspect', 'aspect_cos', 'aspect_sin', 'svf'], errors='ignore')
    # create columns, filled with 0
    df_pts[['elevation', 'slope', 'aspect', 'aspect_cos', 'aspect_sin', 'svf']] = 0

    if method == 'nearest':
        for i, row in df_pts.iterrows():
            d_mini = ds_param.sel(x=row.x, y=row.y, method='nearest')
            df_pts.loc[i, ['elevation', 'slope', 'aspect', 'aspect_cos', 'aspect_sin', 'svf']] = np.array((d_mini.elevation.values,
                                                                                                   d_mini.slope.values,
                                                                                                   d_mini.aspect.values,
                                                                                                   d_mini.aspect_cos,
                                                                                                   d_mini.aspect_sin,
                                                                                                   d_mini.svf.values))
    elif method == 'idw' or method == 'linear':
        for i, row in df_pts.iterrows():
            ind_lat = np.abs(ds_param.y-row.y).argmin()
            ind_lon = np.abs(ds_param.x-row.x).argmin()
            ds_param_pt = ds_param.isel(y=[ind_lat-1, ind_lat, ind_lat+1], x=[ind_lon-1, ind_lon, ind_lon+1])
            Xs, Ys = np.meshgrid(ds_param_pt.x.values, ds_param_pt.y.values)

            if method == 'idw':
                dist = np.sqrt((row.x - Xs)**2 + (row.y - Ys)**2)
                idw = 1/(dist**2)
                weights = idw / np.sum(idw)  # normalize idw to sum(idw) = 1

            if method == 'linear':
                dist = np.sqrt((row.x - Xs)**2 + (row.y - Ys)**2)
                weights = dist / np.sum(dist)
            da_idw = xr.DataArray(data=weights,
                                  coords={
                                    "y": ds_param_pt.y.values,
                                    "x": ds_param_pt.x.values,
                              },
                              dims=["y", "x"]
                              )
            dw = xr.core.weighted.DatasetWeighted(ds_param_pt, da_idw)
            d_mini = dw.sum(['x', 'y'], keep_attrs=True)
            df_pts.loc[i, ['elevation', 'slope', 'aspect', 'aspect_cos', 'aspect_sin', 'svf']] = np.array((d_mini.elevation.values,
                                                                                                   d_mini.slope.values,
                                                                                                   d_mini.aspect.values,
                                                                                                   d_mini.aspect_cos,
                                                                                                   d_mini.aspect_sin,
                                                                                                 d_mini.svf.values))
    else:
        print('ERROR: Method not implemented. Only nearest, linear or idw available')
    return df_pts

def compute_dem_param(dem_file, fname='ds_param.nc', project_directory='./'):
    """
    Function to compute and derive DEM parameters: slope, aspect, sky view factor

    Args:
        dem_file (str): path to raster file (geotif). Raster must be in local cartesian coordinate system (e.g. UTM)

    Returns:  
        dataset: x, y, elev, slope, aspect, svf

    """
    print('\n---> Extracting DEM parameters (slope, aspect, svf)')
    ds = xr.open_rasterio(dem_file).to_dataset('band')
    ds = ds.rename({1: 'elevation'})
    dx = ds.x.diff('x').median().values
    dy = ds.y.diff('y').median().values
    dem_arr = ds.elevation.values
    print('Computing slope and aspect ...')
    slope, aspect = gradient.gradient_d8(dem_arr, dx, dy)
    print('Computing svf ...')
    svf = viewf.viewf(np.double(dem_arr), dx)[0]

    ds['slope'] = (["y", "x"], slope)
    ds['aspect'] = (["y", "x"], np.deg2rad(aspect))
    ds['aspect_cos'] = (["y", "x"], np.cos(np.deg2rad(aspect)))
    ds['aspect_sin'] = (["y", "x"], np.sin(np.deg2rad(aspect)))
    ds['svf'] = (["y", "x"], svf)
    ds.attrs = dict(description="DEM input parameters to TopoSub",
                   author="TopoPyScale, https://github.com/ArcticSnow/TopoPyScale")
    ds.x.attrs = {'units': 'm'}
    ds.y.attrs = {'units': 'm'}
    ds.elevation.attrs = {'units': 'm'}
    ds.slope.attrs = {'units': 'rad'}
    ds.aspect.attrs = {'units': 'rad'}
    ds.aspect_cos.attrs = {'units': 'cosinus'}
    ds.aspect_sin.attrs = {'units': 'sinus'}
    ds.svf.attrs = {'units': 'ratio', 'standard_name': 'svf', 'long_name': 'Sky view factor'}
    te.to_netcdf(ds, fname=f'{project_directory}outputs/{fname}')

    return ds


def compute_horizon(dem_file, azimuth_inc=30, num_threads=None, fname='da_horizon.nc', project_directory='./'):
    """
    Function to compute horizon angles for

    Args:
        dem_file (str): path and filename of the dem
        azimuth_inc (int): angle increment to compute horizons at, in Degrees [0-359]
        num_threads (int): number of threads to parallize on

    Returns: 
        dataarray: all horizon angles for x,y,azimuth coordinates
         
    """
    print('\n---> Computing horizons with {} degree increments'.format(azimuth_inc))
    ds = xr.open_rasterio(dem_file).to_dataset('band')
    ds = ds.rename({1: 'elevation'})
    dx = ds.x.diff('x').median().values

    azimuth = np.arange(-180 + azimuth_inc / 2, 180, azimuth_inc) # center the azimuth in middle of the bin
    arr_val = np.empty((azimuth.shape[0], ds.elevation.shape[0], ds.elevation.shape[1]))

    if num_threads is None:
        pool = ThreadPool(mproc.cpu_count() - 2)
    else:
        pool = ThreadPool(num_threads)

    elev = []
    dxs = []
    for azi in azimuth:
        elev.append(ds.elevation.values)
        dxs.append(dx)

    arr = pool.starmap(horizon.horizon, zip(list(azimuth),
                                                   elev,
                                                   dxs))
    pool.close()
    pool.join()

    for i, a in enumerate(arr):
        arr_val[i,:,:] = a

    da = xr.DataArray(data=np.pi/2 - np.arccos(arr_val),
                      coords={
                          "y": ds.y.values,
                          "x": ds.x.values,
                          "azimuth": azimuth
                      },
                      dims=["azimuth", "y", "x"],
                      name='horizon',
                      attrs={
                          'long_name': 'Horizon angles',
                          'units':'degree'
                      }
                      )
    te.to_netcdf(da.to_dataset(), fname=f'{project_directory}outputs/{fname}')
    return da





"""
Toposcale class definition

S. Filhol, September 2021

project/
    config.ini
    -> input/
        -> dem/
        -> climate/
    -> output/

"""
import glob, os, sys, shutil
import pdb
from pathlib import Path
import re
from munch import DefaultMunch
import pandas as pd
import xarray as xr
import numpy as np
import matplotlib.pyplot as plt
from TopoPyScale import fetch_era5 as fe
from TopoPyScale import topo_param as tp
from TopoPyScale import topo_sub as ts
from TopoPyScale import fetch_dem as fd
from TopoPyScale import solar_geom as sg
from TopoPyScale import topo_scale as ta
from TopoPyScale import topo_export as te
from TopoPyScale import topo_plot as tpl
from TopoPyScale import topo_obs as tpo

from sklearn.preprocessing import StandardScaler


class Topoclass(object):
    """
    A python class to bring the typical use-case of toposcale in a user friendly object
    """

    def __init__(self, config_file):

        try:
            with open(config_file, 'r') as f:
                self.config = DefaultMunch.fromYAML(f)
        except IOError:
            print('ERROR: config file does not exist. Check path.')

        if self.config.outputs.file.clean_outputs:
            # remove outputs directory because if results already exist this causes concat of netcdf files
            try:
                shutil.rmtree(self.config.project.directory + '/outputs/')
                print('---> Output directory cleaned')
            except:
                os.makedirs('/'.join((self.config.project.directory, 'outputs/')))

        # remove output fsm directory
        if self.config.outputs.file.clean_FSM:
            try:
                shutil.rmtree(self.config.project.directory + '/fsm_sims/')
                print('---> FSM directory cleaned')
            except:
                print("---> no FSM directory to clean")

            # remove output fsm directory
            try:
                shutil.rmtree(self.config.project.directory + '/ensemble/')
                print("---> Ensemble directory cleaned")
            except:
                print("---> no ensemble directory to clean")
        # check if tree directory exists. If not create it
        if not os.path.exists('/'.join((self.config.project.directory, 'inputs/'))):
            os.makedirs('/'.join((self.config.project.directory, 'inputs/')))
        if not os.path.exists('/'.join((self.config.project.directory, 'outputs/'))):
            os.makedirs('/'.join((self.config.project.directory, 'outputs/')))

        self.config.climate.path = self.config.project.directory + 'inputs/climate/'
        if not os.path.exists('/'.join((self.config.project.directory, 'inputs/climate/'))):
            os.makedirs('/'.join((self.config.project.directory, 'inputs/climate')))
        if not os.path.exists('/'.join((self.config.project.directory, 'inputs/climate/tmp/'))):
            os.makedirs('/'.join((self.config.project.directory, 'inputs/climate/tmp')))
        if not os.path.exists('/'.join((self.config.project.directory, 'outputs/tmp/'))):
            os.makedirs('/'.join((self.config.project.directory, 'outputs/tmp')))
        if not os.path.exists('/'.join((self.config.project.directory, 'outputs/downscaled'))):
            os.makedirs('/'.join((self.config.project.directory, 'outputs/downscaled')))

        self.config.dem.path = self.config.project.directory + '/inputs/dem/'
        if not os.path.exists('/'.join((self.config.project.directory, 'inputs/dem/'))):
            os.makedirs('/'.join((self.config.project.directory, 'inputs/dem')))

        self.config.dem.filepath = self.config.dem.path + self.config.dem.file
        if not os.path.isfile(self.config.dem.filepath):

            if self.config.project.extent is not None:
                self.config.project.extent = dict(zip(['latN', 'latS', 'lonW', 'lonE'], self.config.project.extent))
            else:
                print('ERROR: no extent provided. Must follow this format: [latN, latS, lonW, lonE]')
                sys.exit()

            fd.fetch_dem(self.config.dem.path, self.config.project.extent, self.config.dem.epsg, self.config.dem.file)
        else:
            print('\n---> DEM file found')

        if self.config.project.extent is not None:
            extent_NSWE = self.config.project.extent
            self.config.project.extent = dict(zip(['latN', 'latS', 'lonW', 'lonE'], extent_NSWE))
        else:
            # little routine extracting lat/lon extent from DEM
            self.config.project.extent = tp.get_extent_latlon(self.config.dem.filepath, self.config.dem.epsg)

        print(self.config.project.extent)
        print('Project lat/lon extent:\n')
        print('\t------------------------------')
        print('\t|        North:{}          |\n\t|West:{}          East:{}|\n\t|        South:{}          |'.format(
            np.round(self.config.project.extent.get('latN'), 1),
            np.round(self.config.project.extent.get('lonW'), 1),
            np.round(self.config.project.extent.get('lonE'), 1),
            np.round(self.config.project.extent.get('latS'), 1)))
        print('\t------------------------------')

        self.toposub = self.Toposub()
        self.plot = Plotting()
        self.ds_solar = None
        self.da_horizon = None

        self.toposub.dem_path = self.config.dem.filepath
        self.toposub.project_directory = self.config.project.directory

        if self.config.project.climate.lower() == 'era5':
            self.get_era5()

        if self.config.project.split.IO:
            self.time_splitter = self.TimeSplitter(self.config.project.start,
                                                   self.config.project.end,
                                                   self.config.project.split.time,
                                                   self.config.outputs.file.ds_solar,
                                                   self.config.outputs.file.downscaled_pt)

    def load_project(self):
        '''
        Function to load pre-existing TopoPyScale project saved in files
        '''
        if os.path.isfile(self.config.project.directory + 'outputs/' + self.config.outputs.file.df_centroids):
            self.toposub.df_centroids = pd.read_pickle(
                self.config.project.directory + 'outputs/' + self.config.outputs.file.df_centroids)
            print(f'---> Centroids file {self.config.outputs.file.df_centroids} exists and loaded')
        else:
            print(f'-> WARNING: Centroid file {self.config.outputs.file.df_centroids} not found')

        if os.path.isfile(self.config.project.directory + 'outputs/' + self.config.outputs.file.ds_param):
            self.toposub.ds_param = xr.open_dataset(
                self.config.project.directory + 'outputs/' + self.config.outputs.file.ds_param)
            print(f'---> DEM parameter file {self.config.outputs.file.ds_param} exists and loaded')
        else:
            print(f'-> WARNING: DEM parameter file {self.config.outputs.file.ds_param} not found')

        if os.path.isfile(self.config.project.directory + 'outputs/' + self.config.outputs.file.ds_solar):
            self.ds_solar = xr.open_dataset(
                self.config.project.directory + 'outputs/' + self.config.outputs.file.ds_solar)
            print(f'---> Solar file {self.config.outputs.file.ds_solar} exists and loaded')
        else:
            print(f'-> WARNING: Solar file {self.config.outputs.file.ds_solar} not found')

        if os.path.isfile(self.config.project.directory + 'outputs/' + self.config.outputs.file.da_horizon):
            self.da_horizon = xr.open_dataarray(
                self.config.project.directory + 'outputs/' + self.config.outputs.file.da_horizon)
            print(f'---> Horizon file {self.config.outputs.file.da_horizon} exists and loaded')
        else:
            print(f'-> WARNING: Horizon file {self.config.outputs.file.da_horizon} not found')

        flist = glob.glob(f'{self.config.project.directory}outputs/downscaled/{self.config.outputs.file.downscaled_pt}')
        if len(flist) > 0:
            print('---> Loading downscaled points \n ...')
            self.downscaled_pts = xr.open_mfdataset(
                f'{self.config.project.directory}outputs/downscaled/{self.config.outputs.file.downscaled_pt}',
                concat_dim='point_id',
                combine='nested',
                parallel=True)
            print(f'---> Downscaled point files {self.config.outputs.file.ds_param} exists and loaded')
        else:
            print(f'-> WARNING: Downscale point files {self.config.outputs.file.downscaled_pt} not found')

    class Toposub:
        """
        Sub-Class to initialize variables to store TopoSub variables
        """

        def __init__(self):
            self.dem_path = None
            self.project_directory = None
            self.ds_param = None
            self.df_centroids = None
            self.kmeans_obj = None
            self.scaler = None

        def plot_clusters_map(self, var='cluster_labels', cmap=plt.cm.hsv, figsize=(14, 10)):
            ts.plot_center_clusters(self.dem_path, self.ds_param, self.df_centroids, var=var, cmap=cmap,
                                    figsize=figsize)

        def write_landform(self):
            ts.write_landform(self.dem_path, self.ds_param, self.project_directory)

    def compute_dem_param(self):
        if os.path.isfile(self.config.project.directory + 'outputs/' + self.config.outputs.file.ds_param):
            self.toposub.ds_param = xr.open_dataset(
                self.config.project.directory + 'outputs/' + self.config.outputs.file.ds_param)
            print(f'---> DEM parameter file {self.config.outputs.file.ds_param} exists and loaded')
        else:
            self.toposub.ds_param = tp.compute_dem_param(self.config.dem.filepath,
                                                         fname=self.config.outputs.file.ds_param,
                                                         project_directory=self.config.project.directory)

    def search_optimum_number_of_clusters(self,
                                          cluster_range=np.arange(100, 1000, 200),
                                          scaler_type=StandardScaler(),
                                          plot=True):
        '''
        Function to test what would be an appropriate number of clusters
        Args:
            cluster_range (int array): numpy array or list of number of clusters to compute scores.

        Returns:

        '''
        print('---> Testing number of clusters')
        print(f'Computing scores for {cluster_range} clusters')
        df_param = ts.ds_to_indexed_dataframe(self.toposub.ds_param)
        print(f'Variables used in clustering: {list(df_param.columns.values)}')

        df_param = ts.ds_to_indexed_dataframe(self.toposub.ds_param)
        df_nclusters = ts.search_number_of_clusters(df_param,
                                                    method=self.config.sampling.toposub.clustering_method,
                                                    cluster_range=cluster_range,
                                                    features=self.config.sampling.toposub.clustering_features,
                                                    scaler_type=scaler_type,
                                                    plot=plot)

        # print to console stats about
        print(df_nclusters)

        return df_nclusters

    def extract_pts_param(self, method='nearest', **kwargs):
        """
        Function to use a list point as input rather than cluster centroids from DEM segmentation (topo_sub.py/self.clustering_dem()).

        Args:
            df (dataFrame):
            method (str): method of sampling
            **kwargs: pd.read_csv() parameters
        Returns:
        """
        self.toposub.df_centroids = pd.read_csv(
            self.config.project.directory + 'inputs/dem/' + self.config.sampling.points.csv_file, **kwargs)
        self.toposub.df_centroids['point_id'] = self.toposub.df_centroids.index.astype(int)
        self.toposub.df_centroids = tp.extract_pts_param(self.toposub.df_centroids, self.toposub.ds_param,
                                                         method=method)

    def extract_topo_cluster_param(self):
        """
        Function to segment a DEM in clusters and retain only the centroids of each cluster.

        :return:

        TODO:
        - try to migrate most code of this function to topo_sub.py as a function itself segment_topo()
        """
        df_param = ts.ds_to_indexed_dataframe(self.toposub.ds_param)
        df_scaled, self.toposub.scaler = ts.scale_df(df_param,
                                                     features=self.config.sampling.toposub.clustering_features)
        if self.config.sampling.toposub.clustering_method.lower() == 'kmean':
            self.toposub.df_centroids, self.toposub.kmeans_obj, df_param['cluster_labels'] = ts.kmeans_clustering(
                df_scaled,
                features=self.config.sampling.toposub.clustering_features,
                n_clusters=self.config.sampling.toposub.n_clusters,
                seed=self.config.sampling.toposub.random_seed)
        elif self.config.sampling.toposub.clustering_method.lower() == 'minibatchkmean':
            self.toposub.df_centroids, self.toposub.kmeans_obj, df_param[
                'cluster_labels'] = ts.minibatch_kmeans_clustering(
                df_scaled,
                n_clusters=self.config.sampling.toposub.n_clusters,
                features=self.config.sampling.toposub.clustering_features,
                n_cores=self.config.project.CPU_cores,
                seed=self.config.sampling.toposub.random_seed)
        else:
            print('ERROR: {} clustering method not available'.format(self.config.sampling.toposub.clustering_method))
        self.toposub.df_centroids = ts.inverse_scale_df(self.toposub.df_centroids, self.toposub.scaler,
                                                        features=self.config.sampling.toposub.clustering_features)

        # logic to add variables not used as clustering predictors into df_centroids
        feature_list = self.config.sampling.toposub.clustering_features.keys()
        flist = list(feature_list)
        flist.append('cluster_labels')
        if len(df_param.columns) > len(flist):
            tmp = df_param.groupby('cluster_labels').mean()
            for var in df_param.drop(flist, axis=1).columns:
                self.toposub.df_centroids[var] = tmp[var]
        self.toposub.df_centroids['point_id'] = self.toposub.df_centroids.index.astype(int)
        self.toposub.ds_param['cluster_labels'] = (
            ["y", "x"], np.reshape(df_param.cluster_labels.values, self.toposub.ds_param.slope.shape))

        # update file
        fname = self.config.project.directory + 'outputs/' + self.config.outputs.file.ds_param
        te.to_netcdf(self.toposub.ds_param, fname=fname)

        # update plotting class variable
        self.plot.ds_param = self.toposub.ds_param

    def extract_topo_param(self):
        """
        Function to select which 
        """
        if os.path.isfile(self.config.project.directory + 'outputs/' + self.config.outputs.file.df_centroids):
            self.toposub.df_centroids = pd.read_pickle(
                self.config.project.directory + 'outputs/' + self.config.outputs.file.df_centroids)
            print(f'---> Centroids file {self.config.outputs.file.df_centroids} exists and loaded')
        else:
            if self.config.sampling.method in ['points', 'point']:
                self.extract_pts_param()
            elif self.config.sampling.method == 'toposub':
                self.extract_topo_cluster_param()
            elif self.config.sampling.method == 'both':

                # implement the case one wann run both toposub and a list of points
                print('ERROR: method not yet implemented')

            else:
                print('ERROR: Extraction method not available')

            # Store dataframe to pickle
            self.toposub.df_centroids.to_pickle(
                self.config.project.directory + 'outputs/' + self.config.outputs.file.df_centroids)
            print(f'---> Centroids file {self.config.outputs.file.df_centroids} saved')

    class TimeSplitter():
        def __init__(self, start, end, period, ds_solar_fname, downscale_fname):
            self.start = start
            self.end = end
            self.period = period

            s = pd.date_range(self.start, self.end, freq=f'{self.period}YS')
            self.start_list = list(s.astype(str))
            self.end_list = list((s - pd.Timedelta('1D')).astype(str))[1:]
            self.end_list.append(end.strftime('%Y-%m-%d'))

            self.ds_solar_flist = self._flist_generator(ds_solar_fname)
            self.downscaled_flist = self._flist_generator(downscale_fname)

        def _flist_generator(self, fname):
            # function to creat list of filename
            fcom = fname.split('.')
            flist = []
            for i, start in enumerate(self.start_list):
                flist.append(f'{fcom[0]}_{start}_{self.end_list[i]}.{fcom[1]}')
            return flist

    def compute_solar_geometry(self):

        if self.config.project.split.IO:
            for i, start in enumerate(self.time_splitter.start_list):
                end = self.time_splitter.end_list[i]
                fname = self.config.project.directory + 'outputs/' + self.time_splitter.ds_solar_flist[i]

                if os.path.isfile(fname):
                    self.ds_solar = xr.open_dataset(fname, chunks='auto', engine='h5netcdf')
                    print(f'---> Solar file {self.time_splitter.ds_solar_flist[i]} exists and loaded')
                else:
                    self.ds_solar = sg.get_solar_geom(self.toposub.df_centroids,
                                                      start,
                                                      end,
                                                      self.config.climate[self.config.project.climate].timestep,
                                                      str(self.config.dem.epsg),
                                                      self.config.project.CPU_cores,
                                                      self.time_splitter.ds_solar_flist[i],
                                                      self.config.project.directory)

        else:
            fname = self.config.project.directory + 'outputs/' + self.config.outputs.file.ds_solar
            if os.path.isfile(fname):
                self.ds_solar = xr.open_dataset(fname, chunks='auto', engine='h5netcdf')
                print(f'---> Solar file {self.config.outputs.file.ds_solar} exists and loaded')
            else:
                self.ds_solar = sg.get_solar_geom(self.toposub.df_centroids,
                                                  self.config.project.start,
                                                  self.config.project.end,
                                                  self.config.climate[self.config.project.climate].timestep,
                                                  str(self.config.dem.epsg),
                                                  self.config.project.CPU_cores,
                                                  self.config.outputs.file.ds_solar,
                                                  self.config.project.directory)

    def compute_horizon(self):
        """
        Function to compute horizon angle and sample values for list of points
        """
        fname = self.config.project.directory + 'outputs/' + self.config.outputs.file.da_horizon
        if os.path.isfile(fname):
            self.da_horizon = xr.open_dataarray(fname, engine='h5netcdf')
            print(f'---> Horizon file {self.config.outputs.file.da_horizon} exists and loaded')
        else:
            self.da_horizon = tp.compute_horizon(self.config.dem.filepath,
                                                 self.config.dem.horizon_increments,
                                                 self.config.project.CPU_cores,
                                                 self.config.outputs.file.da_horizon,
                                                 self.config.project.directory)
        tgt_x = tp.xr.DataArray(self.toposub.df_centroids.x.values, dims="points")
        tgt_y = tp.xr.DataArray(self.toposub.df_centroids.y.values, dims="points")

        # In case horizon angles are computed after clustering
        if self.toposub.df_centroids is not None:
            for az in self.da_horizon.azimuth.values:
                self.toposub.df_centroids['hori_azi_' + str(az)] = self.da_horizon.sel(x=tgt_x,
                                                                                       y=tgt_y,
                                                                                       azimuth=az,
                                                                                       method='nearest').values.flatten()
            # update df_centroid file with horizons angle
            self.toposub.df_centroids.to_pickle(
                self.config.project.directory + 'outputs/' + self.config.outputs.file.df_centroids)
            print(f'---> Centroids file {self.config.outputs.file.df_centroids} updated with horizons')

    def downscale_climate(self):
        downscaled_dir = Path(self.config.project.directory, 'outputs', 'downscaled')
        f_pattern = self.config.outputs.file.downscaled_pt

        if '*' not in f_pattern:
            raise ValueError(
                f'The filepattern for the downscaled files does need to have a * in the name. You provided {f_pattern}')

        if self.config.project.split.IO:
            for i, start in enumerate(self.time_splitter.start_list):
                print()
                end = self.time_splitter.end_list[i]
                fname = self.time_splitter.downscaled_flist[i]

                self.ds_solar = None
                self.ds_solar = xr.open_dataset(
                    self.config.project.directory + 'outputs/' + self.time_splitter.ds_solar_flist[i])

                ta.downscale_climate(self.config.project.directory,
                                     self.toposub.df_centroids,
                                     self.da_horizon,
                                     self.ds_solar,
                                     self.config.dem.epsg,
                                     start,
                                     end,
                                     self.config.toposcale.interpolation_method,
                                     self.config.toposcale.LW_terrain_contribution,
                                     self.config.climate[self.config.project.climate].timestep,
                                     self.config.climate.precip_lapse_rate,
                                     fname,
                                     self.config.project.CPU_cores)

            # Concatenate time-splitted outputs along time-dimension
            # TODO: modify code below to concatenate to be parallelized.
            n_digits = len(str(self.toposub.df_centroids.index.max()))

            # clean directory from files with the same downscaled output file pattern (so they get replaced)
            f_pattern_regex = f_pattern.replace('*', '\d+')
            existing_files = sorted(
                [file for file in downscaled_dir.glob('**/*') if re.match(f_pattern_regex, file.name)])
            for file in existing_files:
                file.unlink()
                print(f'existing file {file.name} removed.')

            for pt_id in self.toposub.df_centroids.point_id.values:
                print(f'Concatenating point {pt_id}')
                num = str(pt_id).zfill(n_digits)
                filename = Path(f_pattern.replace('*', num))
                flist = sorted([file for file in downscaled_dir.glob(f'**/{filename.stem}*')])
                ds_list = [xr.open_dataset(file, engine='h5netcdf') for file in flist]

                fout = Path(downscaled_dir, filename)
                ds = xr.concat(ds_list, dim='time')
                ds.to_netcdf(fout, engine='h5netcdf')
                del ds

            # Delete time slice files.
            for fpat in self.time_splitter.downscaled_flist:
                flist = glob.glob(f'{self.config.project.directory}outputs/downscaled/{fpat}')
                for file in flist:
                    os.remove(file)

        else:
            ta.downscale_climate(self.config.project.directory,
                                 self.toposub.df_centroids,
                                 self.da_horizon,
                                 self.ds_solar,
                                 self.config.dem.epsg,
                                 self.config.project.start,
                                 self.config.project.end,
                                 self.config.toposcale.interpolation_method,
                                 self.config.toposcale.LW_terrain_contribution,
                                 self.config.climate[self.config.project.climate].timestep,
                                 self.config.climate.precip_lapse_rate,
                                 self.config.outputs.file.downscaled_pt,
                                 self.config.project.CPU_cores)

        self.downscaled_pts = ta.read_downscaled(
            self.config.project.directory + 'outputs/downscaled/' + self.config.outputs.file.downscaled_pt)
        # update plotting class variables
        self.plot.ds_down = self.downscaled_pts

    def get_era5(self):
        """
        Funtion to call fetching of ERA5 data
        TODO:
        - merge monthly data into one file (cdo?)- this creates massive slow down!
        """
        lonW = self.config.project.extent.get('lonW') - 0.4
        lonE = self.config.project.extent.get('lonE') + 0.4
        latN = self.config.project.extent.get('latN') + 0.4
        latS = self.config.project.extent.get('latS') - 0.4

        # retreive ERA5 surface data
        fe.retrieve_era5(
            self.config.climate[self.config.project.climate].product,
            self.config.project.start,
            self.config.project.end,
            self.config.climate.path,
            latN, latS, lonE, lonW,
            self.config.climate[self.config.project.climate].timestep,
            self.config.climate[self.config.project.climate].download_threads,
            surf_plev='surf'
        )
        # retrieve era5 plevels
        fe.retrieve_era5(
            self.config.climate[self.config.project.climate].product,
            self.config.project.start,
            self.config.project.end,
            self.config.climate.path,
            latN, latS, lonE, lonW,
            self.config.climate[self.config.project.climate].timestep,
            self.config.climate[self.config.project.climate].download_threads,
            surf_plev='plev',
            plevels=self.config.climate[self.config.project.climate].plevels,
        )

    def get_WMO_observations(self):
        """
        Function to download and parse in-situ data from WMO database
        """
        lonW = self.config.project.extent.get('lonW') - 0.4
        lonE = self.config.project.extent.get('lonE') + 0.4
        latN = self.config.project.extent.get('latN') + 0.4
        latS = self.config.project.extent.get('latS') - 0.4

        df = pd.DataFrame()
        df['dates'] = pd.date_range(self.config.project.start,
                                    self.config.project.end, freq='M')
        df['month'] = df.dates.dt.month
        df['year'] = df.dates.dt.year

        bbox = [latS, lonW, latN, lonE]
        for year in years:
            # API only accept one year request
            tpo.fetch_WMO_insitu_observations(years=year,
                                              months=list(df.month.unique().astype(str)),
                                              bbox=bbox)
        tpo.parse_WMO_insitu_observations()

    def to_cryogrid(self, fname_format='Cryogrid_pt_*.nc', precip_partition='continuous'):
        """
        wrapper function to export toposcale output to cryosgrid format from TopoClass
        
        Args:
            fname_format (str): filename format. point_id is inserted where * is

        """
        path = self.config.project.directory + 'outputs/'
        if 'cluster:labels' in self.toposub.ds_param.keys():
            label_map = True
            da_label = self.toposub.ds_param.cluster_labels
        else:
            label_map = False
            da_label = None
        te.to_cryogrid(self.downscaled_pts,
                       self.toposub.df_centroids,
                       fname_format=fname_format,
                       path=path,
                       label_map=label_map,
                       snow_partition_method=precip_partition,
                       da_label=da_label,
                       climate_dataset_name=self.config.project.climate,
                       project_author=self.config.project.authors)

    def to_fsm(self, fname_format='FSM_pt_*.txt'):
        """
        function to export toposcale output to FSM format
        """
        te.to_fsm(self.downscaled_pts, f'{self.config.project.directory}outputs/' + fname_format)

    def to_crocus(self, fname_format='CROCUS_pt_*.nc', scale_precip=1):
        """
        function to export toposcale output to crocus format .nc. This functions saves one file per point_id
        
        Args:
            fout_format (str): filename format. point_id is inserted where * is
            scale_precip(float): scaling factor to apply on precipitation. Default is 1
        """
        te.to_crocus(self.downscaled_pts,
                     self.toposub.df_centroids,
                     fname_format=f'{self.config.project.directory}outputs/' + fname_format,
                     scale_precip=scale_precip,
                     climate_dataset_name=self.config.project.climate,
                     project_author=self.config.project.authors)

    def to_snowmodel(self, fname_format='Snowmodel_stn_*.csv'):
        """
        function to export toposcale output to snowmodel format .ascii, for single station standard

            fout_format: str, filename format. point_id is inserted where * is
        """
        te.to_micromet_single_station(self.downscaled_pts,
                                      self.toposub.df_centroids,
                                      fname_format=f'{self.config.project.directory}outputs/' + fname_format,
                                      na_values=-9999,
                                      headers=False)

    def to_netcdf(self, file_out='output.nc', variables=None):
        """
        function to export toposcale output to one single generic netcdf format, compressed

        Args:
            file_out (str): name of export file
            variables (list str): list of variable to export. Default exports all variables
        """
        if variables is None:
            variables = list(self.downscaled_pts.keys())

        te.to_netcdf(self.downscaled_pts[variables], f'{self.config.project.directory}outputs/' + file_out, variables)
        print('---> File {} saved'.format(file_out))

    def to_snowpack(self, fname_format='smet_pt_*.smet'):
        """
        function to export toposcale output to FSM format
        """
        te.to_snowpack(self.downscaled_pts, f'{self.config.project.directory}outputs/' + fname_format)

    def to_geotop(self, fname_format='meteo_*.txt'):
        """
        function to export toposcale output to FSM format
        """
        te.to_geotop(self.downscaled_pts, f'{self.config.project.directory}outputs/' + fname_format)

    def to_musa(self,
                fname_met='musa_met.nc',
                fname_labels='musa_labels.nc'):
        """
        function to export TopoPyScale output in a format compatible with MuSa
        MuSa: https://github.com/ealonsogzl/MuSA

        Args:
            fname: filename of the netcdf 
        """
        try:
            labels = self.toposub.ds_param.cluster_labels
        except Exception:
            labels = None

        te.to_musa(ds=self.downscaled_pts,
                   df_pts=self.toposub.df_centroids,
                   da_label=labels,
                   fname_met=fname_met,
                   fname_labels=fname_labels,
                   path=self.config.project.directory + 'outputs/',
                   climate_dataset_name=self.config.project.climate,
                   project_authors=self.config.project.authors
                   )
        print('---> File ready for MuSa {} saved'.format(fname_met))


class Plotting:
    '''
    Sub-Class with plotting functions for topoclass object
    '''

    def __int__(self):
        self.ds_param = None
        self.ds_down = None

    def map_variable(self,
                     var='t',
                     time_step=1,
                     time=None,
                     cmap=plt.cm.RdBu_r,
                     hillshade=True,
                     **kwargs):
        # add logic to check ds_down and ds_param exist.
        tpl.map_variable(self.ds_down,
                         self.ds_param,
                         time_step=time_step,
                         time=time,
                         var=var,
                         cmap=cmap,
                         hillshade=hillshade,
                         **kwargs)

    def map_terrain(self, var='elevation', hillshade=True, **kwargs):
        tpl.map_terrain(self.ds_param,
                        var=var,
                        hillshade=hillshade,
                        **kwargs
                        )

    def map_center_clusters(self,
                            background_var='elevation',
                            cmap=plt.cm.viridis,
                            hillshade=True,
                            **kwargs):
        print('to be implemented')

    def map_clusters(self, **kwargs):
        tp.map_clusters(self.ds_down,
                        self.toposub.ds_param,
                        **kwargs)

    def timeseries(self):
        print('To be implemented')

    def solar_geom(self):
        print('To be implemented')

    def horizon(self):
        print('To be implemented')
